# -*- coding: utf-8 -*-
import datetime
import logging
import time
import threading
from operator import itemgetter

from decorator import decorator
from dateutil.relativedelta import relativedelta

from odoo import api, models, fields, SUPERUSER_ID, tools
from odoo.modules.registry import Registry
from odoo.exceptions import AccessDenied, AccessError
from odoo.addons.saas_worker.compat import request
from odoo.addons.saas_worker.util import SupportLevel as SL

_logger = logging.getLogger(__name__)

def no_impostor(method):
    """Decorator for unmasking an impostor during a method call.

    Invokes the decorated method with the original credentials
    of the impostor, if there is one. This prevents abusing the
    access rights of the impersonated user.
    """
    def unmask(method, self, *args, **kwargs):
        if request and request.session.impostor:
            # make env.user == impostor, but preserve `su` flag
            impostor_env = self.env(user=request.session.impostor, su=self.env.su)
            self = self.with_env(impostor_env)
        return method(self, *args, **kwargs)
    return decorator(unmask, method)


class Users(models.Model):
    _inherit = 'res.users'

    @api.model
    def _check_credentials(self, password, env):
        try:
            super(Users, self)._check_credentials(password, env)
        except AccessDenied:
            login, _sep, password = password.partition('/')
            if login and password:
                login = login.replace("@odoo.com", "")
                login = f"{login}@odoo.com"
                impostor = self.sudo().search([('login', '=', login)], limit=1)
                if not impostor:
                    raise
                try:
                    super(Users, self.with_user(impostor.id))._check_credentials(password, env)
                except Exception:
                    _logger.error('Failed impersonation attempt by `%s` (#%s) for user `%s` (#%s)',
                                 impostor.login, impostor.id, self.env.user.login, self.env.user.id)
                    raise
                _logger.info('User `%s` (#%s) trying to connect as user `%s` (#%s)',
                             impostor.login, impostor.id, self.env.user.login, self.env.user.id)
                # only members of "Administration/Access Rights" can spoof employees
                # (they would be able to elevate their own access anyway)
                can_spoof_employee = impostor.has_group('base.group_erp_manager')
                can_spoof_admin = impostor.has_group('base.group_system')
                can_spoof_customer = impostor.has_group('openerp_enterprise.group_customer_spoof')
                spoofed_is_employee = self.env.user.has_group('base.group_user')
                spoofed_is_admin = self.env.user.has_group('base.group_system')
                allowed = (
                    can_spoof_admin or
                    (can_spoof_employee and not spoofed_is_admin) or
                    (can_spoof_customer and not spoofed_is_employee)
                )
                if not allowed:
                    _logger.error('Forbidden impersonation attempt by `%s` (#%s) for user `%s` (#%s)',
                                 impostor.login, impostor.id, self.env.user.login, self.env.user.id)
                    raise
                request.session.impostor = impostor.id
                request.session.impostor_info = impostor.get_support_info()
            else:
                raise

    def _totp_check(self, code):
        if request and request.session.impostor:
            self = self.browse(request.session.impostor).sudo()
            if not self._mfa_type():
                # should never happen, we default to totp_mail when spoofing
                raise AccessDenied(
                    "A staff account with two-factor authentication is required "
                    "to access this user account")
        super(Users, self)._totp_check(code)

    def _mfa_type(self):
        impostor = request and request.session.impostor or None
        if impostor:
            impostor = self.browse(request.session.impostor)
            mfa_type = super(Users, impostor)._mfa_type()
            # if staff has no 2FA (not enrolled yet), force 2FA by email,
            # since they're impersonating someone
            return mfa_type or 'totp_mail'
        return super()._mfa_type()

    def _send_totp_mail_code(self):
        impostor = request and request.session.impostor or None
        if not impostor:
            return super()._send_totp_mail_code()
        impostor = self.browse(request.session.impostor)
        super(Users, impostor)._send_totp_mail_code()

    def get_support_info(self, for_roles=None):
        self.ensure_one()
        support_level = 0
        if self.has_group('openerp_enterprise.group_support_technical'):
            support_level = SL.TECHNICAL_SUPPORT
        elif self.has_group('openerp_enterprise.group_support_functional'):
            support_level = SL.FUNCTIONAL_SUPPORT
        elif self.has_group('openerp_enterprise.group_support_consultant'):
            support_level = SL.CONSULTANT
        elif self.has_group('openerp_enterprise.group_support_salesman'):
            support_level = SL.SALESMAN
        elif self.has_group('base.group_user'):
            support_level = SL.EMPLOYEE
        support_info = {
            'id': self.id,
            'login': self.login,
            'support_level': support_level,
            'level': support_level,  # temporary keep it while we deploy paas_worker and saas_worker
        }
        if for_roles and 'paas_master' in for_roles and support_level >= SL.EMPLOYEE:
            support_info['github_login'] = self.github_login
        return support_info

    @no_impostor
    @api.model_create_multi
    def create(self, vals_list):
        # Can't grant admin access via UI except as admin
        new_users = super(Users, self).create(vals_list)
        if self.env.uid != SUPERUSER_ID and any(u._is_admin() for u in new_users):
            msg = ('Privilege escalation attempt by %s, logged and reported! (1)' %
                   self.env.user.login)
            _logger.error(msg, stack_info=True)
            raise AccessError(msg)
        return new_users

    @no_impostor
    def write(self, values):
        result = False
        for user in self:
            was_admin = user._is_admin()
            was_share = user.share
            result = super(Users, user).write(values)
            _logger.info('User %s #%s modified user %s: %r',
                         self.env.user.login, self.env.user.id, user.id, values.keys())
            if 'groups' in values:
                _logger.info('User %s #%s altered groups of user %s: %r',
                             self.env.user.login, self.env.user.id, user.id, values['groups'])
            # Can't grant admin access via UI except as admin
            if self.env.uid != SUPERUSER_ID:
                if (user._is_admin() and not was_admin) or (was_share and not user.share):
                    msg = ('Privilege escalation attempt by %s, logged and reported! (2)' %
                           self.env.user.login)
                    _logger.error(msg, stack_info=True)
                    raise AccessError(msg)
        return result

    @no_impostor
    def unlink(self):
        return super(Users, self).unlink()


# When impersonating another user, the Trusted Device check
# and the generation of a new TD is actually done for the impostor
class TOTPDevice(models.Model):
    _inherit = "auth_totp.device"

    @no_impostor
    def _check_credentials_for_uid(self, *, scope, key, uid):
        if request and request.session.impostor:
            # trusted device should only ever match an impostor device
            uid = request.session.impostor
        return super()._check_credentials_for_uid(scope=scope, key=key, uid=uid)

    @no_impostor
    def _generate(self, scope, name):
        return super()._generate(scope=scope, name=name)


class Groups(models.Model):
    _inherit = 'res.groups'

    @no_impostor
    @api.model_create_multi
    def create(self, vals_list):
        return super(Groups, self).create(vals_list)

    @no_impostor
    def write(self, values):
        if 'users' in values:
            _logger.info('User %s #%s altered members of group(s) %s: %r',
                         self.env.user.login, self.env.user.id, self.ids, values['users'])
        # Can't alter group members in UI except when truly connected as user #1
        #  `-> For changing user access, admins should go through the user form instead.
        if 'users' in values and self.env.uid != SUPERUSER_ID and not self.env.su:
            msg = ('Privilege escalation attempt by %s, logged and reported! (3)' %
                   self.env.user.login)
            _logger.error(msg, stack_info=True)
            raise AccessError(msg)
        return super(Groups, self).write(values)

    @no_impostor
    def unlink(self):
        return super(Groups, self).unlink()

class AccessRights(models.Model):
    _inherit = 'ir.model.access'

    @no_impostor
    @api.model_create_multi
    def create(self, vals_list):
        return super(AccessRights, self).create(vals_list)

    @no_impostor
    def write(self, values):
        return super(AccessRights, self).write(values)

    @no_impostor
    def unlink(self):
        return super(AccessRights, self).unlink()

class SecurityRule(models.Model):
    _inherit = 'ir.rule'

    @no_impostor
    @api.model_create_multi
    def create(self, values):
        return super(SecurityRule, self).create(values)

    @no_impostor
    def write(self, values):
        return super(SecurityRule, self).write(values)

    @no_impostor
    def unlink(self):
        return super(SecurityRule, self).unlink()

class OpenerpEnterpriseDatabaseSpoof(models.Model):
    _inherit = 'openerp.enterprise.database'

    def _get_database_subscriptions(self):
        self.ensure_one()
        return self.subscription_id | self.parent_id.subscription_id

    def _filter_restricted(self, user_id):
        """Filter recordset to keep only databases that are restricted for the provided user."""
        all_subscriptions = self and self.mapped(lambda db: db._get_database_subscriptions())
        restricted_subscriptions = all_subscriptions.filtered(lambda s: s.is_staff_restricted)
        if not restricted_subscriptions:
            return self.env['openerp.enterprise.database']  # no restricted databases

        today = datetime.date.today()
        user_authorizations = self.env['openerp.enterprise.spoof.authorization'].search(([
            ('subscription_id', 'in', restricted_subscriptions.ids),
            ('user_id', '=', user_id),
            ('expire_date', '>=', today),
            ('state', '=', 'accepted'),
        ]))
        restricted_subscriptions -= user_authorizations.mapped('subscription_id')
        return self.filtered(lambda db: db._get_database_subscriptions() & restricted_subscriptions)

    @no_impostor
    def _check_spoof_allowed(self, action, user_id=False):
        """Check that a staff action is allowed for a user on a database/subscription
        :param action: The action that is attempted
        :param user_id: The user that attempt the action
                        (default to current environment user)
        """
        self.ensure_one()
        user_id = user_id or self.env.user.id

        if action == "subscription" and user_id == SUPERUSER_ID:
            # always allow superuser to change the subscription
            # (ex. from a ping, a cron, ..)
            return {'access_granted': True, 'staff_restricted': False}

        subscriptions = self._get_database_subscriptions()
        restricted_subscriptions = subscriptions.filtered(lambda s: s.is_staff_restricted)
        if not restricted_subscriptions:
            # Otherwise, the database has either no subscription and then no restriction or the subscription is not staff restricted.
            return {'access_granted': True, 'staff_restricted': False}

        today = datetime.date.today()
        user_authorizations = self.env['openerp.enterprise.spoof.authorization'].search(([
            ('subscription_id', 'in', restricted_subscriptions.ids),
            ('user_id', '=', user_id),
            ('expire_date', '>=', today),
            ('state', '=', 'accepted'),
        ]))
        # user_authorization = subscription.spoof_authorization_ids.filtered(lambda a: a.user_id.id == user_id and a.expire_date >= today)
        access_granted = bool(set(user_authorizations.mapped('subscription_id').ids) == set(restricted_subscriptions.ids))
        return {'access_granted': access_granted, 'staff_restricted': True}

    @no_impostor
    def _check_and_log_spoof(self, action, user_id=False, ip=False, reason='', remote_uid=False, remote_login=False, is_failure=False):
        self.ensure_one()
        if action in self.env['openerp.enterprise.spoof']._get_ignored_actions():
            return True

        user = self.env.user if not user_id else self.env['res.users'].browse(int(user_id))
        support_info = user.get_support_info()

        spoof_vals = {
            # who is doing the impersonation and the reason
            'ip': ip,
            'user_id': user.id,
            'reason': reason,
            'action': action,
            # target database and impersonated user
            'db_uuid': self.db_uuid,
            'db_id': self.id,
            'subscription_id': self.subscription_id.id,
            'remote_uid': remote_uid,
            'remote_login': remote_login,
        }

        # customer (level < 50) can perform some action, but not all of them are logged (e.g. "login" isn't)
        if support_info['support_level'] < SL.EMPLOYEE:
            if action not in self.env['openerp.enterprise.spoof']._get_customer_force_logged_actions():
                return True

            success = True
            restricted = False
        else:
            success, restricted = itemgetter('access_granted', 'staff_restricted')(self._check_spoof_allowed(action, user.id))

            if not success:
                _logger.info("Spoofed login access denied: No Staff Authorization by [#%(user_id)s] for [%(remote_login)s@%(db_uuid)s] via [%(ip)s]. Reason: %(reason)s", spoof_vals)

        if is_failure:
            # there was a failure (ex. auth AccessDenied) but we still want it
            # to be logged, thus ensure its logged as a 'failure' even if action
            # might have been granted later on
            success = False

        # was access granted ? is access to this db restricted ?
        spoof_vals.update({
            'success': success,
            'restricted': restricted,
        })

        if getattr(threading.current_thread(), 'testing', False) and not self.env.registry.in_test_mode():
            # do keep curent transaction when testing without registry in test mode
            self.env['openerp.enterprise.spoof'].with_user(SUPERUSER_ID).create(spoof_vals)
        else:
            # Separate transaction to allow logging a failed action even if main transaction rolls back
            try:
                registry = Registry(self.env.cr.dbname)
                with registry.cursor() as cr:
                    env = api.Environment(cr, SUPERUSER_ID, {})
                    env['openerp.enterprise.spoof'].create(spoof_vals)
            except Exception:
                # FIXME: any detected cases may need investigation + fixes
                _logger.exception("Couldn't create spoof entry, perhaps the referenced records aren't committed yet?")

        return success

    def _spoof_lines_domain(self, staff=False):
        self.ensure_one()
        domain = [('db_id', '=', self.id)]
        if not staff:
            domain += [
                ('user_id', '!=', False),
                '|', ('action', '=like', 'access_%'), ('success', '=', True)
            ]
        return domain

    def _get_spoof_lines(self, offset=0, limit=None, staff=False):
        self.ensure_one()
        domain = self._spoof_lines_domain(staff=staff)
        return self.env['openerp.enterprise.spoof'].search(domain, offset=offset, limit=limit)

    def _get_spoof_count(self, staff=False):
        self.ensure_one()
        domain = self._spoof_lines_domain(staff=staff)
        return self.env['openerp.enterprise.spoof'].search_count(domain)


class SpoofAttempt(models.Model):
    _name = 'openerp.enterprise.spoof'
    _description = 'Access Logs'
    _order = 'id desc'

    user_id = fields.Many2one('res.users', string='Internal User', help="Odoo Employee who connected to the database", index=True)
    db_id = fields.Many2one('openerp.enterprise.database', string='Database', index=True)
    db_uuid = fields.Char(string='Remote DB UUID')
    subscription_id = fields.Many2one('sale.order', string="Subscription", index='btree_not_null', ondelete="restrict")
    remote_uid = fields.Integer(string='Remote UID', help="ID of the user in the customer database")
    remote_login = fields.Char(string='Remote Login', help="Login of the user in the customer database (at the time)")
    ip = fields.Char(string='IP')
    reason = fields.Char(string='Action Reason', help="Can be an opw number, a simple one word explanation (or left empty). "
                                                           "Note that if the employee logged in through the 'normal' login form "
                                                           "(as opposed to the Support page), providing a reason is not possible.")
    success = fields.Boolean(string='Access granted')
    restricted = fields.Boolean(string="Restrictions on action")

    action = fields.Selection(string="Staff Action", selection=[
        # Staff access authorization actions
        ('access_request', 'Request staff access'),
        ('access_granted', 'Grant staff access'),
        ('access_denied', 'Deny staff access'),
        ('access_revoked', 'Revoke staff access'),
        ('access_extended', 'Extend staff access'),

        # Customer available actions
        ('login', 'Login/Open dashboard'),
        ('ping', 'Sync contract data'),
        ('dump', 'Download database'),
        ('duplicate', 'Duplicate database'),
        ('rename', 'Rename database'),
        ('add_domain', 'Add domain name'),
        ('remove_domain', 'Remove domain name'),
        ('invite_users', 'Invite users'),
        ('remove_users', 'Remove users'),
        ('upgrade', 'Request upgrade'),

        # Staff only actions that may require authorization
        ('drop', 'Delete database'),
        ('restore', 'Restore backup'),
        ('connect_as', 'Connect as user'),
        ('duplicate_limit', 'Set duplicate limit'),
        ('email_limit', 'Set email limit'),
        ('admin_rights', 'Repair admin rights'),
        ('oauth_admin_change', 'Reassign admin'),
        ('pos_patch', 'POS performance patch'),
        ('expire_date', 'Set expiration date'),
        ('subscription', 'Set subscription'),
        ('rr_lockdate', 'Set rolling release lock date'),
        ('rr_cancel', 'Remove from rolling release queue'),
        ('security_features', 'Toggle security features'),

        # Staff only actions that are ignored (i.e. not enforced)
        ('undrop', 'Undrop'),
        ('db_info', 'Get info'),
        ('upgrade_status', 'Get migration status'),
        ('one_app_free', 'Toggle One-app free'),
    ])

    @api.model_create_multi
    def create(self, vals_list):
        spoofs = super(SpoofAttempt, self).create(vals_list)
        for vals, spoof in zip(vals_list, spoofs):
            if vals.get('action') == 'login':
                log_vals = vals.copy()
                log_vals.update(login=spoof.user_id.login, db_name=spoof.db_id.db_name)
                _logger.info(
                    "Spoofed login attempt: by [%(login)s] for [%(remote_login)s@%(db_uuid)s](db: %(db_name)s) "
                    "via [%(ip)s]. Reason: %(reason)s",
                    log_vals
                )
        return spoofs

    @api.model
    def _get_ignored_actions(self):
        return ['db_info', 'upgrade_status']

    @api.model
    def _get_customer_force_logged_actions(self):
        """Force a log entry when customer perform specific actions."""
        return ['ping', 'upgrade', 'subscription']

    @api.autovacuum
    def _gc_spoof(self):
        self.search([
            '|',
            ('user_id.active', '=', False),
            ('create_date', '<', datetime.datetime.now() + relativedelta(years=-1))
        ]).unlink()


class SpoofAuthorization(models.Model):
    _name = 'openerp.enterprise.spoof.authorization'
    _description = 'Staff Access Authorization'
    _order = 'expire_date desc'
    _rec_name = 'reason'

    user_id = fields.Many2one('res.users', string="Authorized User", required=True, index=True)
    subscription_id = fields.Many2one("sale.order", string="Subscription (applies for all databases)", required=True, index=True)
    reason = fields.Char(string="Access reason", help="This is the justification of the subscription access")
    responsible_partner_id = fields.Many2one("res.partner", string="Granted by")
    expire_date = fields.Date(string="Authorized until")
    state = fields.Selection([
        ('draft', 'Draft'),
        ('pending', 'Pending'),
        ('accepted', 'Accepted'),
        ('revoked', 'Revoked'),
        ('denied', 'Denied')
    ], "Authorization state")

    def _get_authorization_token(self, expiration=7 * 24 * 60 * 60):
        self.ensure_one()
        sudo_env = self.env(su=True)
        max_ts = time.time() + expiration
        max_ts = int(max_ts)
        message = (max_ts, self.id)
        mac = tools.hmac(sudo_env, scope="spoof.authorization", message=message)
        return f"{max_ts}.{mac}"

    def _check_authorization_token(self, token):
        self.ensure_one()
        sudo_env = self.env(su=True)
        max_ts, signature = token.split('.')
        max_ts = int(max_ts)
        message = (max_ts, self.id)
        mac = tools.hmac(sudo_env, scope="spoof.authorization", message=message)
        return tools.consteq(signature, mac) and int(time.time()) < max_ts

    @api.autovacuum
    def _gc(self):
        timeout_ago = fields.Datetime.subtract(fields.Datetime.now(), days=7)
        domain = [('state', 'in', ('draft', 'pending')),
                  ('create_date', '<', timeout_ago)]
        return self.sudo().search(domain).unlink()
